(*************************************************************************
Copyright (c) 2005-2007, Sergey Bochkanov (ALGLIB project).

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer listed
  in this license in the documentation and/or other materials
  provided with the distribution.

- Neither the name of the copyright holders nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*************************************************************************)
unit lq;
interface
uses Math, Ap, Sysutils, reflections;

procedure LQDecomposition(var A : TReal2DArray;
     M : Integer;
     N : Integer;
     var Tau : TReal1DArray);
procedure UnpackQFromLQ(const A : TReal2DArray;
     M : Integer;
     N : Integer;
     const Tau : TReal1DArray;
     QRows : Integer;
     var Q : TReal2DArray);
procedure LQDecompositionUnpacked(A : TReal2DArray;
     M : Integer;
     N : Integer;
     var L : TReal2DArray;
     var Q : TReal2DArray);

implementation

(*************************************************************************
LQ decomposition of a rectangular matrix of size MxN

Input parameters:
    A   -   matrix A whose indexes range within [1..M, 1..N].
    M   -   number of rows in matrix A.
    N   -   number of columns in matrix A.

Output parameters:
    A   -   matrices L and Q in compact form (see below)
    Tau -   array of scalar factors which are used to form
            matrix Q. Array whose index ranges within [1..Min(M,N)].

Matrix A is represented as A = LQ, where Q is an orthogonal matrix of size
MxM, L - lower triangular (or lower trapezoid) matrix of size M x N.

The elements of matrix L are located on and below the main diagonal of
matrix A. The elements which are located in Tau array and above the main
diagonal of matrix A are used to form matrix Q as follows:

Matrix Q is represented as a product of elementary reflections

Q = H(k)*H(k-1)*...*H(2)*H(1),

where k = min(m,n), and each H(i) is of the form

H(i) = 1 - tau * v * (v^T)

where tau is a scalar stored in Tau[I]; v - real vector,
so that v(1:i-1) = 0, v(i) = 1, v(i+1:n) stored in A(i,i+1:n).

  -- ALGLIB --
     Copyright 2005 by Bochkanov Sergey
*************************************************************************)
procedure LQDecomposition(var A : TReal2DArray;
     M : Integer;
     N : Integer;
     var Tau : TReal1DArray);
var
    WORK : TReal1DArray;
    T : TReal1DArray;
    I : Integer;
    K : Integer;
    NMIP1 : Integer;
    MinMN : Integer;
    MaxMN : Integer;
    Tmp : Double;
    i_ : Integer;
    i1_ : Integer;
begin
    MinMN := Min(M, N);
    MaxMN := Max(M, N);
    SetLength(WORK, M+1);
    SetLength(T, N+1);
    SetLength(TAU, MinMN+1);
    
    //
    // Test the input arguments
    //
    K := Min(M, N);
    I:=1;
    while I<=K do
    begin
        
        //
        // Generate elementary reflector H(i) to annihilate A(i,i+1:n)
        //
        NMIP1 := N-I+1;
        i1_ := (I) - (1);
        for i_ := 1 to NMIP1 do
        begin
            T[i_] := A[I,i_+i1_];
        end;
        GenerateReflection(T, NMIP1, Tmp);
        Tau[I] := Tmp;
        i1_ := (1) - (I);
        for i_ := I to N do
        begin
            A[I,i_] := T[i_+i1_];
        end;
        T[1] := 1;
        if I<N then
        begin
            
            //
            // Apply H(i) to A(i+1:m,i:n) from the right
            //
            ApplyReflectionFromTheRight(A, Tau[I], T, I+1, M, I, N, WORK);
        end;
        Inc(I);
    end;
end;


(*************************************************************************
Partial unpacking of matrix Q from the LQ decomposition of a matrix A

Input parameters:
    A       -   matrices L and Q in compact form.
                Output of LQDecomposition subroutine.
    M       -   number of rows in given matrix A. M>=0.
    N       -   number of columns in given matrix A. N>=0.
    Tau     -   scalar factors which are used to form Q.
                Output of the LQDecomposition subroutine.
    QRows   -   required number of rows in matrix Q. N>=QRows>=0.

Output parameters:
    Q       -   first QRows rows of matrix Q. Array whose indexes range
                within [1..QRows, 1..N]. If QRows=0, the array remains
                unchanged.

  -- ALGLIB --
     Copyright 2005 by Bochkanov Sergey
*************************************************************************)
procedure UnpackQFromLQ(const A : TReal2DArray;
     M : Integer;
     N : Integer;
     const Tau : TReal1DArray;
     QRows : Integer;
     var Q : TReal2DArray);
var
    I : Integer;
    J : Integer;
    K : Integer;
    MinMN : Integer;
    V : TReal1DArray;
    WORK : TReal1DArray;
    VM : Integer;
    i_ : Integer;
    i1_ : Integer;
begin
    Assert(QRows<=N, 'UnpackQFromLQ: QRows>N!');
    if (M=0) or (N=0) or (QRows=0) then
    begin
        Exit;
    end;
    
    //
    // init
    //
    MinMN := Min(M, N);
    K := Min(MinMN, QRows);
    SetLength(Q, QRows+1, N+1);
    SetLength(V, N+1);
    SetLength(WORK, QRows+1);
    I:=1;
    while I<=QRows do
    begin
        J:=1;
        while J<=N do
        begin
            if I=J then
            begin
                Q[I,J] := 1;
            end
            else
            begin
                Q[I,J] := 0;
            end;
            Inc(J);
        end;
        Inc(I);
    end;
    
    //
    // unpack Q
    //
    I:=K;
    while I>=1 do
    begin
        
        //
        // Apply H(i)
        //
        VM := N-I+1;
        i1_ := (I) - (1);
        for i_ := 1 to VM do
        begin
            V[i_] := A[I,i_+i1_];
        end;
        V[1] := 1;
        ApplyReflectionFromTheRight(Q, Tau[I], V, 1, QRows, I, N, WORK);
        Dec(I);
    end;
end;


(*************************************************************************
LQ decomposition of a rectangular matrix of size MxN

It uses LQDecomposition. L and Q are not output in compact form, but as
separate general matrices. L is filled up by zeros in their corresponding
positions, and Q is generated as a product of elementary reflections.

  -- ALGLIB --
     Copyright 2005 by Bochkanov Sergey
*************************************************************************)
procedure LQDecompositionUnpacked(A : TReal2DArray;
     M : Integer;
     N : Integer;
     var L : TReal2DArray;
     var Q : TReal2DArray);
var
    I : Integer;
    J : Integer;
    Tau : TReal1DArray;
begin
    A := DynamicArrayCopy(A);
    if N<=0 then
    begin
        Exit;
    end;
    SetLength(Q, N+1, N+1);
    SetLength(L, M+1, N+1);
    
    //
    // LQDecomposition
    //
    LQDecomposition(A, M, N, Tau);
    
    //
    // L
    //
    I:=1;
    while I<=M do
    begin
        J:=1;
        while J<=N do
        begin
            if J>I then
            begin
                L[I,J] := 0;
            end
            else
            begin
                L[I,J] := A[I,J];
            end;
            Inc(J);
        end;
        Inc(I);
    end;
    
    //
    // Q
    //
    UnpackQFromLQ(A, M, N, Tau, N, Q);
end;


end.