(*************************************************************************
Copyright (c) 1992-2007 The University of Tennessee.  All rights reserved.

Contributors:
    * Sergey Bochkanov (ALGLIB project). Translation from FORTRAN to
      pseudocode.

See subroutines comments for additional copyrights.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer listed
  in this license in the documentation and/or other materials
  provided with the distribution.

- Neither the name of the copyright holders nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*************************************************************************)
unit qr;
interface
uses Math, Ap, Sysutils, reflections;

procedure QRDecomposition(var A : TReal2DArray;
     M : Integer;
     N : Integer;
     var Tau : TReal1DArray);
procedure UnpackQFromQR(const A : TReal2DArray;
     M : Integer;
     N : Integer;
     const Tau : TReal1DArray;
     QColumns : Integer;
     var Q : TReal2DArray);
procedure QRDecompositionUnpacked(A : TReal2DArray;
     M : Integer;
     N : Integer;
     var Q : TReal2DArray;
     var R : TReal2DArray);

implementation

(*************************************************************************
QR decomposition of a rectangular matrix of size MxN

Input parameters:
    A   -   matrix A whose indexes range within [1..M, 1..N].
    M   -   number of rows in matrix A.
    N   -   number of columns in matrix A.

Output parameters:
    A   -   matrices Q and R in compact form (see below).
    Tau -   array of scalar factors which are used to form
            matrix Q. Array whose index ranges within [1..Min(M,N)].

Matrix A is represented as A = QR, where Q is an orthogonal matrix of size
MxM, R - upper triangular (or upper trapezoid) matrix of size M x N.

The elements of matrix R are located on and above the main diagonal of
matrix A. The elements which are located in Tau array and below the main
diagonal of matrix A are used to form matrix Q as follows:

Matrix Q is represented as a product of elementary reflections

Q = H(1)*H(2)*...*H(k),

where k = min(m,n), and each H(i) is in the form

H(i) = 1 - tau * v * (v^T)

where tau is a scalar stored in Tau[I]; v - real vector,
so that v(1:i-1) = 0, v(i) = 1, v(i+1:m) stored in A(i+1:m,i).

  -- LAPACK routine (version 3.0) --
     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
     Courant Institute, Argonne National Lab, and Rice University
     February 29, 1992
*************************************************************************)
procedure QRDecomposition(var A : TReal2DArray;
     M : Integer;
     N : Integer;
     var Tau : TReal1DArray);
var
    WORK : TReal1DArray;
    T : TReal1DArray;
    I : Integer;
    K : Integer;
    MMIP1 : Integer;
    MinMN : Integer;
    Tmp : Double;
    i_ : Integer;
    i1_ : Integer;
begin
    MinMN := Min(M, N);
    SetLength(WORK, N+1);
    SetLength(T, M+1);
    SetLength(TAU, MinMN+1);
    
    //
    // Test the input arguments
    //
    K := Min(M, N);
    I:=1;
    while I<=K do
    begin
        
        //
        // Generate elementary reflector H(i) to annihilate A(i+1:m,i)
        //
        MMIP1 := M-I+1;
        i1_ := (I) - (1);
        for i_ := 1 to MMIP1 do
        begin
            T[i_] := A[i_+i1_,I];
        end;
        GenerateReflection(T, MMIP1, Tmp);
        Tau[I] := Tmp;
        i1_ := (1) - (I);
        for i_ := I to M do
        begin
            A[i_,I] := T[i_+i1_];
        end;
        T[1] := 1;
        if I<N then
        begin
            
            //
            // Apply H(i) to A(i:m,i+1:n) from the left
            //
            ApplyReflectionFromTheLeft(A, Tau[I], T, I, M, I+1, N, WORK);
        end;
        Inc(I);
    end;
end;


(*************************************************************************
Partial unpacking of matrix Q from the QR decomposition of a matrix A

Input parameters:
    A       -   matrices Q and R in compact form.
                Output of QRDecomposition subroutine.
    M       -   number of rows in given matrix A. M>=0.
    N       -   number of columns in given matrix A. N>=0.
    Tau     -   scalar factors which are used to form Q.
                Output of the QRDecomposition subroutine.
    QColumns -  required number of columns of matrix Q. M>=QColumns>=0.

Output parameters:
    Q       -   first QColumns columns of matrix Q.
                Array whose indexes range within [1..M, 1..QColumns].
                If QColumns=0, the array remains unchanged.

  -- ALGLIB --
     Copyright 2005 by Bochkanov Sergey
*************************************************************************)
procedure UnpackQFromQR(const A : TReal2DArray;
     M : Integer;
     N : Integer;
     const Tau : TReal1DArray;
     QColumns : Integer;
     var Q : TReal2DArray);
var
    I : Integer;
    J : Integer;
    K : Integer;
    MinMN : Integer;
    V : TReal1DArray;
    WORK : TReal1DArray;
    VM : Integer;
    i_ : Integer;
    i1_ : Integer;
begin
    Assert(QColumns<=M, 'UnpackQFromQR: QColumns>M!');
    if (M=0) or (N=0) or (QColumns=0) then
    begin
        Exit;
    end;
    
    //
    // init
    //
    MinMN := Min(M, N);
    K := Min(MinMN, QColumns);
    SetLength(Q, M+1, QColumns+1);
    SetLength(V, M+1);
    SetLength(WORK, QColumns+1);
    I:=1;
    while I<=M do
    begin
        J:=1;
        while J<=QColumns do
        begin
            if I=J then
            begin
                Q[I,J] := 1;
            end
            else
            begin
                Q[I,J] := 0;
            end;
            Inc(J);
        end;
        Inc(I);
    end;
    
    //
    // unpack Q
    //
    I:=K;
    while I>=1 do
    begin
        
        //
        // Apply H(i)
        //
        VM := M-I+1;
        i1_ := (I) - (1);
        for i_ := 1 to VM do
        begin
            V[i_] := A[i_+i1_,I];
        end;
        V[1] := 1;
        ApplyReflectionFromTheLeft(Q, Tau[I], V, I, M, 1, QColumns, WORK);
        Dec(I);
    end;
end;


(*************************************************************************
QR decomposition of a rectangular matrix of size MxN

It uses QRDecomposition. Q and R are not output in compact form, but as
separate general matrices. R is filled up by zeros in their corresponding
positions, and Q is generated as a product of elementary reflections.

  -- ALGLIB --
     Copyright 2005 by Bochkanov Sergey
*************************************************************************)
procedure QRDecompositionUnpacked(A : TReal2DArray;
     M : Integer;
     N : Integer;
     var Q : TReal2DArray;
     var R : TReal2DArray);
var
    I : Integer;
    K : Integer;
    Tau : TReal1DArray;
    WORK : TReal1DArray;
    V : TReal1DArray;
    i_ : Integer;
begin
    A := DynamicArrayCopy(A);
    K := Min(M, N);
    if N<=0 then
    begin
        Exit;
    end;
    SetLength(WORK, M+1);
    SetLength(V, M+1);
    SetLength(Q, M+1, M+1);
    SetLength(R, M+1, N+1);
    
    //
    // QRDecomposition
    //
    QRDecomposition(A, M, N, Tau);
    
    //
    // R
    //
    I:=1;
    while I<=N do
    begin
        R[1,I] := 0;
        Inc(I);
    end;
    I:=2;
    while I<=M do
    begin
        for i_ := 1 to N do
        begin
            R[I,i_] := R[1,i_];
        end;
        Inc(I);
    end;
    I:=1;
    while I<=K do
    begin
        for i_ := I to N do
        begin
            R[I,i_] := A[I,i_];
        end;
        Inc(I);
    end;
    
    //
    // Q
    //
    UnpackQFromQR(A, M, N, Tau, M, Q);
end;


end.