(*************************************************************************
Copyright (c) 2005-2007, Sergey Bochkanov (ALGLIB project).

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer listed
  in this license in the documentation and/or other materials
  provided with the distribution.

- Neither the name of the copyright holders nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*************************************************************************)
unit svd;
interface
uses Math, Ap, Sysutils, reflections, bidiagonal, qr, lq, blas, rotations, bdsvd;

function SVDDecomposition(A : TReal2DArray;
     M : Integer;
     N : Integer;
     UNeeded : Integer;
     VTNeeded : Integer;
     AdditionalMemory : Integer;
     var W : TReal1DArray;
     var U : TReal2DArray;
     var VT : TReal2DArray):Boolean;

implementation

(*************************************************************************
Singular value decomposition of a rectangular matrix.

The algorithm calculates the singular value decomposition of a matrix of
size MxN: A = U * S * V^T

The algorithm finds the singular values and, optionally, matrices U and V^T.
The algorithm can find both first min(M,N) columns of matrix U and rows of
matrix V^T (singular vectors), and matrices U and V^T wholly (of sizes MxM
and NxN respectively).

Take into account that the subroutine does not return matrix V but V^T.

Input parameters:
    A           -   matrix to be decomposed.
                    Array whose indexes range within [1..M, 1..N].
    M           -   number of rows in matrix A.
    N           -   number of columns in matrix A.
    UNeeded     -   0, 1 or 2. See the description of the parameter U.
    VTNeeded    -   0, 1 or 2. See the description of the parameter VT.
    AdditionalMemory -
                    If the parameter:
                     * equals 0, the algorithm doesnt use additional
                       memory (lower requirements, lower performance).
                     * equals 1, the algorithm uses additional
                       memory of size min(M,N)*min(M,N) of real numbers.
                       It often speeds up the algorithm.
                     * equals 2, the algorithm uses additional
                       memory of size M*min(M,N) of real numbers.
                       It allows to get a maximum performance.
                    The recommended value of the parameter is 2.

Output parameters:
    W           -   contains singular values in descending order.
    U           -   if UNeeded=0, U isn't changed, the left singular vectors
                    are not calculated.
                    if Uneeded=1, U contains left singular vectors (first
                    min(M,N) columns of matrix U). Array whose indexes range
                    within [1..M, 1..Min(M,N)].
                    if UNeeded=2, U contains matrix U wholly. Array whose
                    indexes range within [1..M, 1..M].
    VT          -   if VTNeeded=0, VT isnt changed, the right singular vectors
                    are not calculated.
                    if VTNeeded=1, VT contains right singular vectors (first
                    min(M,N) rows of matrix V^T). Array whose indexes range
                    within [1..min(M,N), 1..N].
                    if VTNeeded=2, VT contains matrix V^T wholly. Array whose
                    indexes range within [1..N, 1..N].

  -- ALGLIB --
     Copyright 2005 by Bochkanov Sergey
*************************************************************************)
function SVDDecomposition(A : TReal2DArray;
     M : Integer;
     N : Integer;
     UNeeded : Integer;
     VTNeeded : Integer;
     AdditionalMemory : Integer;
     var W : TReal1DArray;
     var U : TReal2DArray;
     var VT : TReal2DArray):Boolean;
var
    TauQ : TReal1DArray;
    TauP : TReal1DArray;
    Tau : TReal1DArray;
    E : TReal1DArray;
    WORK : TReal1DArray;
    T2 : TReal2DArray;
    IsUpper : Boolean;
    MinMN : Integer;
    NCU : Integer;
    NRVT : Integer;
    NRU : Integer;
    NCVT : Integer;
    I : Integer;
    J : Integer;
    IM1 : Integer;
    SM : Double;
begin
    A := DynamicArrayCopy(A);
    Result := True;
    if (M=0) or (N=0) then
    begin
        Exit;
    end;
    Assert((UNeeded>=0) and (UNeeded<=2), 'SVDDecomposition: wrong parameters!');
    Assert((VTNeeded>=0) and (VTNeeded<=2), 'SVDDecomposition: wrong parameters!');
    Assert((AdditionalMemory>=0) and (AdditionalMemory<=2), 'SVDDecomposition: wrong parameters!');
    
    //
    // initialize
    //
    MinMN := Min(M, N);
    SetLength(W, MinMN+1);
    NCU := 0;
    NRU := 0;
    if UNeeded=1 then
    begin
        NRU := M;
        NCU := MinMN;
        SetLength(U, NRU+1, NCU+1);
    end;
    if UNeeded=2 then
    begin
        NRU := M;
        NCU := M;
        SetLength(U, NRU+1, NCU+1);
    end;
    NRVT := 0;
    NCVT := 0;
    if VTNeeded=1 then
    begin
        NRVT := MinMN;
        NCVT := N;
        SetLength(VT, NRVT+1, NCVT+1);
    end;
    if VTNeeded=2 then
    begin
        NRVT := N;
        NCVT := N;
        SetLength(VT, NRVT+1, NCVT+1);
    end;
    
    //
    // M much larger than N
    // Use bidiagonal reduction with QR-decomposition
    //
    if M>1.6*N then
    begin
        if UNeeded=0 then
        begin
            
            //
            // No left singular vectors to be computed
            //
            QRDecomposition(A, M, N, Tau);
            I:=2;
            while I<=N do
            begin
                J:=1;
                while J<=I-1 do
                begin
                    A[I,J] := 0;
                    Inc(J);
                end;
                Inc(I);
            end;
            ToBidiagonal(A, N, N, TauQ, TauP);
            UnpackPTFromBidiagonal(A, N, N, TauP, NRVT, VT);
            UnpackDiagonalsFromBidiagonal(A, N, N, IsUpper, W, E);
            Result := BidiagonalSVDDecomposition(W, E, N, IsUpper, False, U, 0, A, 0, VT, NCVT);
            Exit;
        end
        else
        begin
            
            //
            // Left singular vectors (may be full matrix U) to be computed
            //
            QRDecomposition(A, M, N, Tau);
            UnpackQFromQR(A, M, N, Tau, NCU, U);
            I:=2;
            while I<=N do
            begin
                J:=1;
                while J<=I-1 do
                begin
                    A[I,J] := 0;
                    Inc(J);
                end;
                Inc(I);
            end;
            ToBidiagonal(A, N, N, TauQ, TauP);
            UnpackPTFromBidiagonal(A, N, N, TauP, NRVT, VT);
            UnpackDiagonalsFromBidiagonal(A, N, N, IsUpper, W, E);
            if AdditionalMemory<1 then
            begin
                
                //
                // No additional memory can be used
                //
                MultiplyByQFromBidiagonal(A, N, N, TauQ, U, M, N, True, False);
                Result := BidiagonalSVDDecomposition(W, E, N, IsUpper, False, U, M, A, 0, VT, NCVT);
            end
            else
            begin
                
                //
                // Large U. Transforming intermediate matrix T2
                //
                SetLength(WORK, Max(M, N)+1);
                UnpackQFromBidiagonal(A, N, N, TauQ, N, T2);
                CopyMatrix(U, 1, M, 1, N, A, 1, M, 1, N);
                InplaceTranspose(T2, 1, N, 1, N, WORK);
                Result := BidiagonalSVDDecomposition(W, E, N, IsUpper, False, U, 0, T2, N, VT, NCVT);
                MatrixMatrixMultiply(A, 1, M, 1, N, False, T2, 1, N, 1, N, True, 1.0, U, 1, M, 1, N, 0.0, WORK);
            end;
            Exit;
        end;
    end;
    
    //
    // N much larger than M
    // Use bidiagonal reduction with LQ-decomposition
    //
    if N>1.6*M then
    begin
        if VTNeeded=0 then
        begin
            
            //
            // No right singular vectors to be computed
            //
            LQDecomposition(A, M, N, Tau);
            I:=1;
            while I<=M-1 do
            begin
                J:=I+1;
                while J<=M do
                begin
                    A[I,J] := 0;
                    Inc(J);
                end;
                Inc(I);
            end;
            ToBidiagonal(A, M, M, TauQ, TauP);
            UnpackQFromBidiagonal(A, M, M, TauQ, NCU, U);
            UnpackDiagonalsFromBidiagonal(A, M, M, IsUpper, W, E);
            SetLength(WORK, M+1);
            InplaceTranspose(U, 1, NRU, 1, NCU, WORK);
            Result := BidiagonalSVDDecomposition(W, E, M, IsUpper, False, A, 0, U, NRU, VT, 0);
            InplaceTranspose(U, 1, NRU, 1, NCU, WORK);
            Exit;
        end
        else
        begin
            
            //
            // Right singular vectors (may be full matrix VT) to be computed
            //
            LQDecomposition(A, M, N, Tau);
            UnpackQFromLQ(A, M, N, Tau, NRVT, VT);
            I:=1;
            while I<=M-1 do
            begin
                J:=I+1;
                while J<=M do
                begin
                    A[I,J] := 0;
                    Inc(J);
                end;
                Inc(I);
            end;
            ToBidiagonal(A, M, M, TauQ, TauP);
            UnpackQFromBidiagonal(A, M, M, TauQ, NCU, U);
            UnpackDiagonalsFromBidiagonal(A, M, M, IsUpper, W, E);
            SetLength(WORK, Max(M, N)+1);
            InplaceTranspose(U, 1, NRU, 1, NCU, WORK);
            if AdditionalMemory<1 then
            begin
                
                //
                // No additional memory available
                //
                MultiplyByPFromBidiagonal(A, M, M, TauP, VT, M, N, False, True);
                Result := BidiagonalSVDDecomposition(W, E, M, IsUpper, False, A, 0, U, NRU, VT, N);
            end
            else
            begin
                
                //
                // Large VT. Transforming intermediate matrix T2
                //
                UnpackPTFromBidiagonal(A, M, M, TauP, M, T2);
                Result := BidiagonalSVDDecomposition(W, E, M, IsUpper, False, A, 0, U, NRU, T2, M);
                CopyMatrix(VT, 1, M, 1, N, A, 1, M, 1, N);
                MatrixMatrixMultiply(T2, 1, M, 1, M, False, A, 1, M, 1, N, False, 1.0, VT, 1, M, 1, N, 0.0, WORK);
            end;
            InplaceTranspose(U, 1, NRU, 1, NCU, WORK);
            Exit;
        end;
    end;
    
    //
    // M<=N
    // We can use inplace transposition of U to get rid of columnwise operations
    //
    if M<=N then
    begin
        ToBidiagonal(A, M, N, TauQ, TauP);
        UnpackQFromBidiagonal(A, M, N, TauQ, NCU, U);
        UnpackPTFromBidiagonal(A, M, N, TauP, NRVT, VT);
        UnpackDiagonalsFromBidiagonal(A, M, N, IsUpper, W, E);
        SetLength(WORK, M+1);
        InplaceTranspose(U, 1, NRU, 1, NCU, WORK);
        Result := BidiagonalSVDDecomposition(W, E, MinMN, IsUpper, False, A, 0, U, NRU, VT, NCVT);
        InplaceTranspose(U, 1, NRU, 1, NCU, WORK);
        Exit;
    end;
    
    //
    // Simple bidiagonal reduction
    //
    ToBidiagonal(A, M, N, TauQ, TauP);
    UnpackQFromBidiagonal(A, M, N, TauQ, NCU, U);
    UnpackPTFromBidiagonal(A, M, N, TauP, NRVT, VT);
    UnpackDiagonalsFromBidiagonal(A, M, N, IsUpper, W, E);
    if (AdditionalMemory<2) or (UNeeded=0) then
    begin
        
        //
        // We cant use additional memory or there is no need in such operations
        //
        Result := BidiagonalSVDDecomposition(W, E, MinMN, IsUpper, False, U, NRU, A, 0, VT, NCVT);
    end
    else
    begin
        
        //
        // We can use additional memory
        //
        SetLength(T2, MinMN+1, M+1);
        CopyAndTranspose(U, 1, M, 1, MinMN, T2, 1, MinMN, 1, M);
        Result := BidiagonalSVDDecomposition(W, E, MinMN, IsUpper, False, U, 0, T2, M, VT, NCVT);
        CopyAndTranspose(T2, 1, MinMN, 1, M, U, 1, M, 1, MinMN);
    end;
end;


end.
